# -*- coding: utf-8 -*-

import pytest
import os
import shutil

from video_compressor.adapters.ffmpeg import ffmpegAdapter
from video_compressor.compressor import compressToTargetSize
from video_compressor.exceptions import MissingLibraryError, InvalidVideoInput

__author__ = "Lenselle Nicolas"
__copyright__ = "Lenselle Nicolas"
__license__ = "mit"

# remove sound on video
# check if video has sound or not


@pytest.fixture(scope="function")
def tempdir():
    return os.getcwd()+'/tmp/'


@pytest.fixture(scope="function")
def tempfile(tempdir):
    def path(filename=''):
        return tempdir+filename
    return path


@pytest.fixture(scope="function", autouse=True)
def beforeAll(tempdir):
    if os.path.exists(tempdir):
        shutil.rmtree(tempdir)
    os.mkdir(tempdir)
    yield
    shutil.rmtree(tempdir)


def test_error_with_invalid_ffmpeg():
    with pytest.raises(MissingLibraryError):
        ffmpegAdapter(input='./src/tests/sample.mp4', bin_ffmpeg='ffmpeg-custom-bin')
    with pytest.raises(MissingLibraryError):
        ffmpegAdapter(input='./src/tests/sample.mp4').bin_ffmpeg('ffmpeg-custom-bin')


def test_error_with_invalid_input():
    with pytest.raises(InvalidVideoInput):
        ffmpegAdapter(input='./src/tests/corrupted-sample.mp4')

    with pytest.raises(InvalidVideoInput):
        ffmpegAdapter(input='./src/tests/unexisting-sample.mp4')


def test_error_with_invalid_ouput():
    pass


def test_error_with_invalid_mute_value():
    pass


def test_error_with_invalid_scale_value():
    pass


def test_get_video_audio():
    assert ffmpegAdapter(input='./src/tests/sample.mp4').volumedetect() is False
    assert ffmpegAdapter(input='./src/tests/mute-sample.mp4').volumedetect() is True


def test_get_video_bitrate():
    assert ffmpegAdapter(input='./src/tests/sample.mp4').get_video_bitrate() == 2200634


def test_get_audio_bitrate():
    assert ffmpegAdapter(input='./src/tests/sample.mp4').get_audio_bitrate() == 133274


def test_get_video_resolution():
    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    w, h = video.get_resolution()
    assert h == 540
    assert w == 960


def test_get_video_duration():
    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    d = video.get_duration()
    assert d == 4871.533


def test_get_video_size():
    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    s = video.get_size()
    assert s == 1507453


def test_mute_a_video(tempfile):
    video = ffmpegAdapter(input='./src/tests/sample.mp4')

    sample_copy = tempfile(filename='sample-copy.mp4')
    sample_muted = tempfile(filename='sample-muted.mp4')

    video.export(sample_copy)
    video.mute(True).export(sample_muted)

    assert ffmpegAdapter(input=sample_copy).volumedetect() is False
    assert ffmpegAdapter(input=sample_muted).volumedetect() is True


def test_scale_video(tempfile):

    sample54x96 = tempfile(filename='sample-54x96.mp4')

    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    w, h = video.get_resolution()
    assert w == 960
    assert h == 540

    video.scale(96, 54).export(sample54x96)
    w, h = ffmpegAdapter(input=sample54x96).get_resolution()
    assert w == 96
    assert h == 54


def test_scale_video_keep_ratio(tempfile):

    sample640 = tempfile(filename='sample-54x96.mp4')

    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    w1, h1 = video.get_resolution()

    video.scale(640, -1).export(sample640)
    w2, h2 = ffmpegAdapter(input=sample640).get_resolution()
    assert w2 // h2 == w1 // h1


def test_reduce_video_bitrate(tempfile):

    sample1Mbs = tempfile(filename='sample-1mbs.mp4')
    video = ffmpegAdapter(input='./src/tests/sample.mp4')

    video.bitrate(1000000).export(sample1Mbs)
    assert ffmpegAdapter(input=sample1Mbs).get_video_bitrate() < 1000000


def test_reduce_video_to_specific_size(tempfile):

    # 1 507 453 is ./src/tests/sample.mp4 original size

    maxSize = 0.5 * 1000 * 1000 * 8  # 0.5MB in bits
    sample500k = tempfile(filename='sample-500k.mp4')

    video = ffmpegAdapter(input='./src/tests/sample.mp4')
    compressToTargetSize(video, maxSize, sample500k)

    assert ffmpegAdapter(sample500k).get_size() < (maxSize / 8)

# target_size=$(( 25 * 1000 * 1000 * 8 )) # 25MB in bits
# length=`ffprobe -v error -show_entries format=duration -of default=noprint_wrappers=1:nokey=1 input.mp4`
# length_round_up=$(( ${length%.*} + 1 ))
# total_bitrate=$(( $target_size / $length_round_up ))
# audio_bitrate=$(( 128 * 1000 )) # 128k bit rate
# video_bitrate=$(( $total_bitrate - $audio_bitrate ))
# ffmpeg -i input.mp4 -b:v $video_bitrate -maxrate:v $video_bitrate -bufsize:v $(( $target_size / 20 )) -b:a $audio_bitrate output.mp4



from shutil import which
import subprocess
import os

from ..exceptions import MissingLibraryError, InvalidVideoInput


def process(command):
    process = subprocess.run(command, shell=True, stdout=subprocess.PIPE)
    return (
        (process.stdout or b'').decode("utf-8"),
        (process.stderr or b'').decode("utf-8")
    )


class ffprobeCmdBuilder():

    def __init__(
        self,
        input,
        **options
    ):
        self.input = input

    @property
    def ffprobe(self):
        return "ffprobe -v error -select_streams v:0 -of csv=s=,:p=0"

    @property
    def resolution(self):
        return f"{self.ffprobe} -show_entries stream=width,height {self.input}"

    @property
    def video_bitrate(self):
        return f"{self.ffprobe} -show_entries stream=bit_rate {self.input}"

    @property
    def audio_bitrate(self):
        return f"{self.ffprobe} -select_streams a:0 -show_entries stream=bit_rate {self.input}"

    @property
    def duration(self):
        return f"{self.ffprobe} -show_entries stream=duration {self.input}"


class ffmpegCmdBuilder():

    def __init__(
        self,
        input=None,
        bin_ffmpeg=None,
        mute=None,
        scale=None,
        bitrate=None,
        output=None,
        **options
    ):
        self.input = input
        self.bin_ffmpeg = bin_ffmpeg
        self.mute = mute
        self.scale = scale
        self.bitrate = bitrate
        self.output = output

    @property
    def ffmpeg(self):
        return f"{self.bin_ffmpeg} -i {self.input} -q:v 0"

    @property
    def mutefilter(self):
        return '-an' if self.mute else ''

    @property
    def scalefilter(self):
        return f'-vf scale={self.scale[0]}:{self.scale[1]}' if self.scale else ''

    @property
    def bitratefilter(self):
        return f'-maxrate:v {self.bitrate}' if self.bitrate else ''

    @property
    def pipestdout(self):
        return '-f null /dev/null 2>&1'

    @property
    def integrity(self):
        return f'{self.ffmpeg} -v error {self.pipestdout}'

    @property
    def export(self):
        return f'{self.ffmpeg} {self.mutefilter} {self.scalefilter} {self.bitratefilter} {self.output}'

    @property
    def volumedetect(self):
        return f"{self.ffmpeg} -af 'volumedetect' {self.pipestdout}"


class ffmpegAdapter():

    def __init__(
        self,
        input=None,
        bin_ffmpeg='ffmpeg',
        mute=None,
        scale=None,
        bitrate=None,
        output=None
    ):
        self._bin_ffmpeg = bin_ffmpeg
        self._input = input
        self._mute = mute
        self._scale = scale
        self._bitrate = bitrate
        self._output = output

        self.ffmpeg = ffmpegCmdBuilder(**self.options())
        self.ffprobe = ffprobeCmdBuilder(**self.options())

        self._check_bin_validity(bin_ffmpeg)
        self._check_input_integrity(input)

    def options(self):
        return {
            'input': self._input,
            'bin_ffmpeg': self._bin_ffmpeg,
            'mute': self._mute,
            'scale': self._scale,
            'output': self._output,
            'bitrate': self._bitrate
        }

    def update(self, **updates):
        options = self.options()
        options.update(updates)
        return ffmpegAdapter(**options)

    def bin_ffmpeg(self, bin_ffmpeg):
        self._check_bin_validity(bin_ffmpeg)
        return self.update(bin_ffmpeg=bin_ffmpeg)

    def input(self, input):
        return self.update(input=input)

    def output(self, output):
        return self.update(output=output)

    def mute(self, mute):
        return self.update(mute=mute)

    def scale(self, *scale):
        return self.update(scale=scale)

    def bitrate(self, bitrate):
        return self.update(bitrate=bitrate)

    def _check_bin_validity(self, bin):
        if which(bin) is None:
            raise MissingLibraryError

    def _check_input_integrity(self, input):
        trace, error = process(self.ffmpeg.integrity)
        if 'Invalid data' in trace or 'No such file or directory' in trace:
            raise InvalidVideoInput(trace)

    def export(self, output):
        process(self.output(output).ffmpeg.export)

    def volumedetect(self):
        volumetrace, error = process(self.ffmpeg.volumedetect)
        return 'Parsed_volumedetect' not in volumetrace

    def get_resolution(self):
        resolution, error = process(self.ffprobe.resolution)
        return map(int, resolution.split(','))

    def get_video_bitrate(self):
        bitrate, error = process(self.ffprobe.video_bitrate)
        return int(bitrate) if bitrate else 0

    def get_audio_bitrate(self):
        bitrate, error = process(self.ffprobe.audio_bitrate)
        return int(bitrate) if bitrate else 0

    def get_duration(self):
        duration, error = process(self.ffprobe.duration)
        duration = duration.replace('\n', '')
        return float(duration) * 1000

    def get_size(self):
        return os.path.getsize(self._input)





def compressToTargetSize(video, targetSize, output):

    length = video.get_duration() / 1000
    total_bitrate = targetSize / (length+1)
    audio_bitrate = video.get_audio_bitrate()
    video_bitrate = total_bitrate - audio_bitrate

    video.bitrate(video_bitrate).export(output)

